@@ -59,6 +59,14 @@ unexpected_excep(int vector)
     }
 }
 
+extern void __divide_error(void);
+
+void
+__divide_error(void)
+{
+    return;
+}
+
 enum {
     ARCH_PREP = 0,
     ARCH_MAC99,
@@ -68,7 +76,7 @@ enum {
 
 int is_apple(void)
 {
-    return 1;
+    return is_oldworld() || is_newworld();
 }
 
 int is_oldworld(void)
@@ -87,12 +95,12 @@ static const pci_arch_t known_arch[] = {
         .name = "PREP",
         .vendor_id = PCI_VENDOR_ID_MOTOROLA,
         .device_id = PCI_DEVICE_ID_MOTOROLA_RAVEN,
-        .cfg_addr = 0x80800000,
-        .cfg_data = 0x800c0000,
+        .cfg_addr = 0x80000cf8,
+        .cfg_data = 0x80000cfc,
         .cfg_base = 0x80000000,
         .cfg_len = 0x00100000,
-        .host_mem_base = 0xf0000000,
-        .pci_mem_base = 0xf0000000,
+        .host_pci_base = 0xc0000000,
+        .pci_mem_base = 0x100000, /* avoid VGA at 0xa0000 */
         .mem_len = 0x10000000,
         .io_base = 0x80000000,
         .io_len = 0x00010000,
@@ -108,7 +116,7 @@ static const pci_arch_t known_arch[] = {
         .cfg_data = 0xf2c00000,
         .cfg_base = 0xf2000000,
         .cfg_len = 0x02000000,
-        .host_mem_base = 0x80000000,
+        .host_pci_base = 0x0,
         .pci_mem_base = 0x80000000,
         .mem_len = 0x10000000,
         .io_base = 0xf2000000,
@@ -125,7 +133,7 @@ static const pci_arch_t known_arch[] = {
         .cfg_data = 0xf0c00000,
         .cfg_base = 0xf0000000,
         .cfg_len = 0x02000000,
-        .host_mem_base = 0x80000000,
+        .host_pci_base = 0x0,
         .pci_mem_base = 0x80000000,
         .mem_len = 0x10000000,
         .io_base = 0xf2000000,
@@ -142,7 +150,7 @@ static const pci_arch_t known_arch[] = {
         .cfg_data = 0xfee00000,
         .cfg_base = 0x80000000,
         .cfg_len = 0x7f000000,
-        .host_mem_base = 0x80000000,
+        .host_pci_base = 0x0,
         .pci_mem_base = 0x80000000,
         .mem_len = 0x10000000,
         .io_base = 0xfe000000,
@@ -255,7 +263,7 @@ cpu_generic_init(const struct cpudef *cpu)
     push_str("timebase-frequency");
     fword("property");
 
-    PUSH(fw_cfg_read_i32(FW_CFG_PPC_CPUFREQ));
+    PUSH(fw_cfg_read_i32(FW_CFG_PPC_CLOCKFREQ));
     fword("encode-int");
     push_str("clock-frequency");
     fword("property");
@@ -613,6 +621,22 @@ static void kvm_of_init(void)
     fword("finish-device");
 }
 
+/*
+ *  filll        ( addr bytes quad -- )
+ */
+
+static void ffilll(void)
+{
+    const u32 longval = POP();
+    u32 bytes = POP();
+    u32 *laddr = (u32 *)cell2pointer(POP());
+    u32 len;
+    
+    for (len = 0; len < bytes / sizeof(u32); len++) {
+        *laddr++ = longval;
+    }   
+}
+
 void
 arch_of_init(void)
 {
@@ -720,6 +744,9 @@ arch_of_init(void)
         push_str("MacRISC");
         fword("encode-string");
         fword("encode+");
+        push_str("MacRISC2");
+        fword("encode-string");
+        fword("encode+");
         push_str("Power Macintosh");
         fword("encode-string");
         fword("encode+");
@@ -831,7 +858,7 @@ arch_of_init(void)
     }
 
     /* Setup default boot devices */
-    snprintf(buf, sizeof(buf), "%s:,\\\\:tbxi %s:,\\ppc\\bootinfo.txt", boot_path, boot_path);
+    snprintf(buf, sizeof(buf), "%s:,\\\\:tbxi %s:,\\ppc\\bootinfo.txt %s:,%%BOOT", boot_path, boot_path, boot_path);
     push_str(buf);
     fword("encode-string");
     push_str("boot-device");
@@ -842,18 +869,6 @@ arch_of_init(void)
     push_str("/chosen");
     fword("find-device");
 
-    push_str(stdin_path);
-    fword("open-dev");
-    fword("encode-int");
-    push_str("stdin");
-    fword("property");
-
-    push_str(stdout_path);
-    fword("open-dev");
-    fword("encode-int");
-    push_str("stdout");
-    fword("property");
-
     push_str(stdin_path);
     fword("pathres-resolve-aliases");
     push_str("input-device");
@@ -864,12 +879,6 @@ arch_of_init(void)
     push_str("output-device");
     fword("$setenv");
 
-    push_str(stdin_path);
-    fword("input");
-
-    push_str(stdout_path);
-    fword("output");
-
 #if 0
     if(getbool("tty-interface?") == 1)
 #endif
@@ -877,6 +886,9 @@ arch_of_init(void)
 
     device_end();
 
+    /* Implementation of filll word (required by BootX) */
+    bind_func("filll", ffilll);
+    
     bind_func("platform-boot", boot);
     bind_func("(go)", go);
 }
